#!/usr/bin/python

# pip install lxml

import sys
import os
import json
import xml.etree.ElementTree as ET
import glob
from tqdm import tqdm

START_BOUNDING_BOX_ID = 1
PRE_DEFINE_CATEGORIES = None
# If necessary, pre-define category and its id
#  PRE_DEFINE_CATEGORIES = {"aeroplane": 1, "bicycle": 2, "bird": 3, "boat": 4,
#  "bottle":5, "bus": 6, "car": 7, "cat": 8, "chair": 9,
#  "cow": 10, "diningtable": 11, "dog": 12, "horse": 13,
#  "motorbike": 14, "person": 15, "pottedplant": 16,
#  "sheep": 17, "sofa": 18, "train": 19, "tvmonitor": 20}


def get(root, name):
    vars = root.findall(name)
    return vars


def get_and_check(root, name, length):
    vars = root.findall(name)
    if len(vars) == 0:
        raise ValueError("Can not find %s in %s." % (name, root.tag))
    if length > 0 and len(vars) != length:
        raise ValueError("The size of %s is supposed to be %d, but is %d." %
                         (name, length, len(vars)))
    if length == 1:
        vars = vars[0]
    return vars


def get_filename_as_int(filename):
    try:
        filename = filename.replace("\\", "/")
        filename = os.path.splitext(os.path.basename(filename))[0]
        return int(filename)
    except:
        raise ValueError("Filename %s is supposed to be an integer." %
                         (filename))

class image_id_create:
    def __init__(self):
        self.image_id_map = dict()
        self.all_images = list()
    def add_image(self, file_name):
        isin =  file_name in self.image_id_map
        if not isin:
            self.all_images.append(file_name)
            int_label = self.all_images.index(file_name)
            self.image_id_map[file_name] = int_label
    def get_filename_as_int(self, filename):
        return self.image_id_map[filename]

def get_categories(xml_files):
    """Generate category name to id mapping from a list of xml files.
    
    Arguments:
        xml_files {list} -- A list of xml file paths.
    
    Returns:
        dict -- category name to id mapping.
    """
    classes_names = []
    for xml_file in xml_files:
        tree = ET.parse(xml_file)
        root = tree.getroot()
        for member in root.findall("object"):
            classes_names.append(member[0].text)
    classes_names = list(set(classes_names))
    classes_names.sort()
    return {name: i for i, name in enumerate(classes_names)}


def convert(xml_dir, json_file):
    xml_files = glob.glob(os.path.join(xml_dir, "*.xml"))
    # If you want to do train/test split, you can pass a subset of xml files to convert function.
    print("Number of xml files: {}".format(len(xml_files)))
    
    
    
    json_dict = {
        "images": [],
        "type": "instances",
        "annotations": [],
        "categories": []
    }
    iic = image_id_create()
    
    if PRE_DEFINE_CATEGORIES is not None:
        categories = PRE_DEFINE_CATEGORIES
    else:
        categories = get_categories(xml_files)
    bnd_id = START_BOUNDING_BOX_ID
    for xml_file in tqdm(xml_files):
        tree = ET.parse(xml_file)
        root = tree.getroot()
        path = get(root, "path")
        if len(path) == 1:
            filename = os.path.basename(path[0].text)
        elif len(path) == 0:
            filename = get_and_check(root, "filename", 1).text
        else:
            raise ValueError("%d paths found in %s" % (len(path), xml_file))
        
        # Modified image id function
        iic.add_image(filename)
        image_id = iic.get_filename_as_int(filename)
#         ## The filename must be a number
#         image_id = get_filename_as_int(filename)
        size = get_and_check(root, "size", 1)
        width = int(get_and_check(size, "width", 1).text)
        height = int(get_and_check(size, "height", 1).text)
        image = {
            "file_name": filename,
            "height": height,
            "width": width,
            "id": image_id,
        }
        json_dict["images"].append(image)
        ## Currently we do not support segmentation.
        #  segmented = get_and_check(root, 'segmented', 1).text
        #  assert segmented == '0'
        for obj in get(root, "object"):
            category = get_and_check(obj, "name", 1).text
            if category not in categories:
                new_id = len(categories)
                categories[category] = new_id
            category_id = categories[category]
            bndbox = get_and_check(obj, "bndbox", 1)
            xmin = int(float(get_and_check(bndbox, "xmin", 1).text)) - 1
            ymin = int(float(get_and_check(bndbox, "ymin", 1).text)) - 1
            xmax = int(float(get_and_check(bndbox, "xmax", 1).text))
            ymax = int(float(get_and_check(bndbox, "ymax", 1).text))
            assert xmax > xmin
            assert ymax > ymin
            o_width = abs(xmax - xmin)
            o_height = abs(ymax - ymin)
            ann = {
                "area": o_width * o_height,
                "iscrowd": 0,
                "image_id": image_id,
                "bbox": [xmin, ymin, o_width, o_height],
                "category_id": category_id,
                "id": bnd_id,
                "ignore": 0,
                "segmentation": [],
            }
            json_dict["annotations"].append(ann)
            bnd_id = bnd_id + 1

    for cate, cid in categories.items():
        cat = {"supercategory": "none", "id": cid, "name": cate}
        json_dict["categories"].append(cat)

    os.makedirs(os.path.dirname(json_file), exist_ok=True)
    json_fp = open(json_file, "w")
    json_str = json.dumps(json_dict)
    json_fp.write(json_str)
    json_fp.close()
    print("Success: {}".format(json_file))

    
# usage
# where train is the folder contains xml files, the output json will be under train/ dir
# convert('train', 'train/train_annotations.json')


import os
from collections import Counter
import nltk
from tqdm import tqdm
import torch


def intersection_over_union(boxes_preds, boxes_labels, box_format="midpoint"):
    """
    Calculates intersection over union
    Parameters:
        boxes_preds (tensor): Predictions of Bounding Boxes (BATCH_SIZE, 4)
        boxes_labels (tensor): Correct Labels of Boxes (BATCH_SIZE, 4)
        box_format (str): midpoint/corners, if boxes (x,y,w,h) or (x1,y1,x2,y2)
    Returns:
        tensor: Intersection over union for all examples
    """

    # Slicing idx:idx+1 in order to keep tensor dimensionality
    # Doing ... in indexing if there would be additional dimensions
    # Like for Yolo algorithm which would have (N, S, S, 4) in shape
    if box_format == "midpoint":

        box1_x1 = boxes_preds[..., 0:1] - boxes_preds[..., 2:3] / 2
        box1_y1 = boxes_preds[..., 1:2] - boxes_preds[..., 3:4] / 2
        box1_x2 = boxes_preds[..., 0:1] + boxes_preds[..., 2:3] / 2
        box1_y2 = boxes_preds[..., 1:2] + boxes_preds[..., 3:4] / 2
        box2_x1 = boxes_labels[..., 0:1] - boxes_labels[..., 2:3] / 2
        box2_y1 = boxes_labels[..., 1:2] - boxes_labels[..., 3:4] / 2
        box2_x2 = boxes_labels[..., 0:1] + boxes_labels[..., 2:3] / 2
        box2_y2 = boxes_labels[..., 1:2] + boxes_labels[..., 3:4] / 2

    elif box_format == "corners":
        box1_x1 = boxes_preds[..., 0:1]
        box1_y1 = boxes_preds[..., 1:2]
        box1_x2 = boxes_preds[..., 2:3]
        box1_y2 = boxes_preds[..., 3:4]
        box2_x1 = boxes_labels[..., 0:1]
        box2_y1 = boxes_labels[..., 1:2]
        box2_x2 = boxes_labels[..., 2:3]
        box2_y2 = boxes_labels[..., 3:4]

    x1 = torch.max(box1_x1, box2_x1)
    y1 = torch.max(box1_y1, box2_y1)
    x2 = torch.min(box1_x2, box2_x2)
    y2 = torch.min(box1_y2, box2_y2)

    # Need clamp(0) in case they do not intersect, then we want intersection to be 0
    intersection = (x2 - x1).clamp(0) * (y2 - y1).clamp(0)
    box1_area = abs((box1_x2 - box1_x1) * (box1_y2 - box1_y1))
    box2_area = abs((box2_x2 - box2_x1) * (box2_y2 - box2_y1))

    return intersection / (box1_area + box2_area - intersection + 1e-6)


def mean_average_precision(
        pred_boxes, true_boxes, iou_threshold=0.5, box_format="midpoint"
):
    """
    Calculates mean average precision
    Parameters:
        pred_boxes (list): list of lists containing all bboxes with each bboxes
        specified as [train_idx, class_prediction, prob_score, x1, y1, x2, y2]
        true_boxes (list): Similar as pred_boxes except all the correct ones
        iou_threshold (float): threshold where predicted bboxes is correct
        box_format (str): "midpoint" or "corners" used to specify bboxes
    Returns:
        float: mAP value across all classes given a specific IoU threshold
    """
    all_classes = list(set([l[1] for l in true_boxes]))

    label_wise_precision = nltk.defaultdict(float)
    label_wise_recall = nltk.defaultdict(float)
    label_wise_ap = nltk.defaultdict(float)

    # list storing all AP for respective classes
    average_precisions = []

    # used for numerical stability later on
    epsilon = 1e-6

    for c in tqdm(all_classes):
        detections = []
        ground_truths = []

        # Go through all predictions and targets,
        # and only add the ones that belong to the
        # current class c
        for detection in pred_boxes:
            if detection[1] == c:
                detections.append(detection)

        for true_box in true_boxes:
            if true_box[1] == c:
                ground_truths.append(true_box)

        # find the amount of bboxes for each training example
        # Counter here finds how many ground truth bboxes we get
        # for each training example, so let's say img 0 has 3,
        # img 1 has 5 then we will obtain a dictionary with:
        # amount_bboxes = {0:3, 1:5}
        amount_bboxes = Counter([gt[0] for gt in ground_truths])

        # We then go through each key, val in this dictionary
        # and convert to the following (w.r.t same example):
        # ammount_bboxes = {0:torch.tensor[0,0,0], 1:torch.tensor[0,0,0,0,0]}
        for key, val in amount_bboxes.items():
            amount_bboxes[key] = torch.zeros(val)

        # sort by box probabilities which is index 2
        detections.sort(key=lambda x: x[2], reverse=True)
        TP = torch.zeros((len(detections)))
        FP = torch.zeros((len(detections)))
        total_true_bboxes = len(ground_truths)

        # If none exists for this class then we can safely skip
        if total_true_bboxes == 0:
            continue
        for detection_idx, detection in enumerate(detections):
            # Only take out the ground_truths that have the same
            # training idx as detection
            ground_truth_img = [
                bbox for bbox in ground_truths if bbox[0] == detection[0]
            ]
            num_gts = len(ground_truth_img)
            best_iou = 0

            for idx, gt in enumerate(ground_truth_img):
                iou = intersection_over_union(
                    torch.tensor(detection[3:]),
                    torch.tensor(gt[3:]),
                    box_format=box_format,
                )

                if iou > best_iou:
                    best_iou = iou
                    best_gt_idx = idx

            if best_iou > iou_threshold:
                # only detect ground truth detection once
                if amount_bboxes[detection[0]][best_gt_idx] == 0:
                    # true positive and add this bounding box to seen
                    TP[detection_idx] = 1
                    amount_bboxes[detection[0]][best_gt_idx] = 1
                else:
                    FP[detection_idx] = 1

            # if IOU is lower then the detection is a false positive
            else:
                FP[detection_idx] = 1

        TP_cumsum = torch.cumsum(TP, dim=0)
        FP_cumsum = torch.cumsum(FP, dim=0)
        recalls = TP_cumsum / (total_true_bboxes + epsilon)
        precisions = TP_cumsum / (TP_cumsum + FP_cumsum + epsilon)
        precisions = torch.cat((torch.tensor([1]), precisions))
        recalls = torch.cat((torch.tensor([0]), recalls))
        # torch.trapz for numerical integration
        label_wise_precision[c] = precisions.numpy()[-1]
        label_wise_recall[c] = recalls.numpy()[-1]
        label_ap = float(torch.trapz(precisions, recalls))
        # average_precisions.append(label_ap)
        label_wise_ap[c] = label_ap

    return dict(label_wise_precision), dict(label_wise_recall), dict(label_wise_ap)


class label_map:
    def __init__(self, pred_txt_dir, truth_txt_dir):
        pred_boxes = []
        true_boxes = []

        """
        expect pred txt sequence:
        each txt contains rows like this:

        8 0.529688 0.829167 0.078125 0.0444444 0.893442
        8 0.723828 0.816667 0.103906 0.0694444 0.905974
        8 0.95 0.778472 0.0921875 0.0708333 0.911328
        0 0.103125 0.111806 0.110937 0.118056 0.935599

        with 

        int_class x y w h conf
        """

        for root, dirs, files in os.walk(pred_txt_dir):
            for file in files:
                if file.endswith(".txt"):
                    the_path = os.path.join(root, file)
                    if 'checkpoint' not in the_path:
                        file_name = the_path.split(os.sep)[-1].split('.')[0]
                        with open(the_path) as f:
                            lines = f.read().splitlines()
                        for l in lines:
                            line_split = l.split(' ')
                            line_split.insert(0, file_name)
                            conf = line_split[-1]
                            line_split.insert(2, conf)
                            line_split = line_split[:-1]
                            line_split[1] = int(line_split[1])
                            line_split[2] = float(line_split[2])
                            line_split[3] = float(line_split[3])
                            line_split[4] = float(line_split[4])
                            line_split[5] = float(line_split[5])
                            line_split[6] = float(line_split[6])
                            pred_boxes.append(line_split)

        """
        expect true txt sequence:
        each txt contains rows like this:

        0 0.104297 0.107639 0.110156 0.120833
        8 0.724219 0.818056 0.098437 0.069444
        8 0.949219 0.784028 0.089063 0.081944
        8 0.530078 0.831250 0.075781 0.048611

        with 

        int_class x y w h
        """

        for root, dirs, files in os.walk(truth_txt_dir):
            for file in files:
                if file.endswith(".txt"):
                    the_path = os.path.join(root, file)
                    if 'checkpoint' not in the_path:
                        file_name = the_path.split(os.sep)[-1].split('.')[0]
                        with open(the_path) as f:
                            lines = f.read().splitlines()
                        for l in lines:
                            line_split = l.split(' ')
                            line_split.insert(0, file_name)
                            line_split.insert(2, 1)
                            line_split[1] = int(line_split[1])
                            line_split[2] = float(line_split[2])
                            line_split[3] = float(line_split[3])
                            line_split[4] = float(line_split[4])
                            line_split[5] = float(line_split[5])
                            line_split[6] = float(line_split[6])
                            true_boxes.append(line_split)

        self.pred_boxes = pred_boxes
        self.true_boxes = true_boxes

    def get_metrics(self, iou_threshold=0.5):
        label_wise_precision, label_wise_recall, label_wise_ap = mean_average_precision(
            self.pred_boxes, self.true_boxes, iou_threshold=iou_threshold, box_format="midpoint"
        )
        return {'precision': label_wise_precision, 'recall': label_wise_recall, 'average precision': label_wise_ap}


    def find_confuse_labels(self, iou_threshold=0.1, pred_boxes=None, true_boxes=None, box_format='midpoint'):
        """
        This will give an idea of how model confused between each labels

        @param pred_boxes: result from label_map class, list of boxes, if None, use self.pred_boxes
        @param true_boxes: result from label_map class, list of boxes, if None, use self.true_boxes
        @param iou_threshold:
        @param box_format:
        @return:
        """
        # do suggest put iou_threshold low since this is not a calculation for metrics

        """
        the iou_threshold still needed think of the case that, there is a true box in the image, but the true box doesn't 
        overlap any preidction box, which mean it is a missed object, so we want to capture this case. If we don't have an
        iou_threshold, simply sorted by ious, then the true box actually doesn't match the sorted top predict box
    
        """
        if pred_boxes is None:
            pred_boxes = self.pred_boxes
        if true_boxes is None:
            true_boxes = self.true_boxes

        label_wise_confuse = nltk.defaultdict(lambda: nltk.defaultdict(int))
        all_img = list(set([i[0] for i in true_boxes]))
        for img in tqdm(all_img):
            img_pred = [i for i in pred_boxes if i[0] == img]
            img_gt = [i for i in true_boxes if i[0] == img]
            for gt in img_gt:
                img_ious = []
                for det in img_pred:
                    iou = intersection_over_union(
                        torch.tensor(det[3:]),
                        torch.tensor(gt[3:]),
                        box_format=box_format,
                    )
                    if iou >= iou_threshold:
                        img_ious.append([iou, det])
                if img_ious:
                    match_pred_box = sorted(img_ious, key=lambda x: x[0], reverse=True)[0][1]
                    det_cls = match_pred_box[1]
                else:
                    det_cls = 'backgroud'

                true_cls = gt[1]
                label_wise_confuse[true_cls][det_cls] += 1
        result = dict()
        for label, confuse_dict in label_wise_confuse.items():
            result[label] = [(k, v) for k, v in sorted(confuse_dict.items(), key=lambda item: item[1], reverse=True)]

        return result


def demo():
    pred_txt_dir = 'yolo_cls/runs/detect/exp/labels'
    truth_txt_dir = 'linmao_camera_data_real/labels/test'
    lm = label_map(pred_txt_dir, truth_txt_dir)
    lm.get_metrics(iou_threshold=0.5)
    lm.find_confuse_labels(iou_threshold=0.1, box_format='midpoint')